"""bump-my-version Command line interface."""

from pathlib import Path
from typing import List, Optional

import questionary
import rich_click as click
from click.core import Context
from tomlkit import dumps

from bumpversion import __version__
from bumpversion.bump import do_bump
from bumpversion.config import get_configuration
from bumpversion.config.create import create_configuration
from bumpversion.config.files import find_config_file
from bumpversion.context import get_context
from bumpversion.files import ConfiguredFile, modify_files
from bumpversion.show import do_show
from bumpversion.ui import get_indented_logger, print_info, setup_logging
from bumpversion.utils import get_overrides
from bumpversion.visualize import visualize

logger = get_indented_logger(__name__)


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
    add_help_option=True,
)
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx: Context) -> None:
    """Version bump your Python project."""
    pass


click.rich_click.OPTION_GROUPS = {
    "bumpversion bump": [
        {
            "name": "Configuration",
            "options": [
                "--config-file",
                "--current-version",
                "--new-version",
                "--parse",
                "--serialize",
                "--search",
                "--replace",
                "--no-configured-files",
                "--ignore-missing-files",
                "--ignore-missing-version",
            ],
        },
        {
            "name": "Output",
            "options": ["--dry-run", "--verbose"],
        },
        {
            "name": "Committing and tagging",
            "options": [
                "--allow-dirty",
                "--commit",
                "--commit-args",
                "--message",
                "--tag",
                "--tag-name",
                "--tag-message",
                "--sign-tags",
            ],
        },
    ]
}


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("args", nargs=-1, type=str)
@click.option(
    "--config-file",
    metavar="FILE",
    required=False,
    envvar="BUMPVERSION_CONFIG_FILE",
    type=click.Path(exists=True),
    help="Config file to read most of the variables from.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    required=False,
    envvar="BUMPVERSION_VERBOSE",
    help="Print verbose logging to stderr. Can specify several times for more verbosity.",
)
@click.option(
    "--allow-dirty/--no-allow-dirty",
    default=None,
    required=False,
    envvar="BUMPVERSION_ALLOW_DIRTY",
    help="Don't abort if working directory is dirty, or explicitly abort if dirty.",
)
@click.option(
    "--current-version",
    metavar="VERSION",
    required=False,
    envvar="BUMPVERSION_CURRENT_VERSION",
    help="Version that needs to be updated",
)
@click.option(
    "--new-version",
    metavar="VERSION",
    required=False,
    envvar="BUMPVERSION_NEW_VERSION",
    help="New version that should be in the files",
)
@click.option(
    "--parse",
    metavar="REGEX",
    required=False,
    envvar="BUMPVERSION_PARSE",
    help="Regex parsing the version string",
)
@click.option(
    "--serialize",
    metavar="FORMAT",
    multiple=True,
    required=False,
    envvar="BUMPVERSION_SERIALIZE",
    help="How to format what is parsed back to a version",
)
@click.option(
    "--search",
    metavar="SEARCH",
    required=False,
    envvar="BUMPVERSION_SEARCH",
    help="Template for complete string to search",
)
@click.option(
    "--replace",
    metavar="REPLACE",
    required=False,
    envvar="BUMPVERSION_REPLACE",
    help="Template for complete string to replace",
)
@click.option(
    "--regex/--no-regex",
    default=None,
    envvar="BUMPVERSION_REGEX",
    help="Treat the search parameter as a regular expression or explicitly do not treat it as a regular expression.",
)
@click.option(
    "--no-configured-files",
    is_flag=True,
    envvar="BUMPVERSION_NO_CONFIGURED_FILES",
    help=(
        "Only replace the version in files specified on the command line, "
        "ignoring the files from the configuration file."
    ),
)
@click.option(
    "--ignore-missing-files/--no-ignore-missing-files",
    default=None,
    envvar="BUMPVERSION_IGNORE_MISSING_FILES",
    help="Ignore any missing files when searching and replacing in files.",
)
@click.option(
    "--ignore-missing-version/--no-ignore-missing-version",
    default=None,
    envvar="BUMPVERSION_IGNORE_MISSING_VERSION",
    help="Ignore any Version Not Found errors when searching and replacing in files.",
)
@click.option(
    "--dry-run",
    "-n",
    is_flag=True,
    envvar="BUMPVERSION_DRY_RUN",
    help="Don't write any files, just pretend.",
)
@click.option(
    "--commit/--no-commit",
    default=None,
    envvar="BUMPVERSION_COMMIT",
    help="Commit to version control",
)
@click.option(
    "--tag/--no-tag",
    default=None,
    envvar="BUMPVERSION_TAG",
    help="Create a tag in version control",
)
@click.option(
    "--sign-tags/--no-sign-tags",
    default=None,
    envvar="BUMPVERSION_SIGN_TAGS",
    help="Sign tags if created",
)
@click.option(
    "--tag-name",
    metavar="TAG_NAME",
    required=False,
    envvar="BUMPVERSION_TAG_NAME",
    help="Tag name (only works with --tag)",
)
@click.option(
    "--tag-message",
    metavar="TAG_MESSAGE",
    required=False,
    envvar="BUMPVERSION_TAG_MESSAGE",
    help="Tag message",
)
@click.option(
    "-m",
    "--message",
    metavar="COMMIT_MSG",
    required=False,
    envvar="BUMPVERSION_MESSAGE",
    help="Commit message",
)
@click.option(
    "--commit-args",
    metavar="COMMIT_ARGS",
    required=False,
    envvar="BUMPVERSION_COMMIT_ARGS",
    help="Extra arguments to commit command",
)
def bump(
    # version_part: str,
    args: list,
    config_file: Optional[str],
    verbose: int,
    allow_dirty: Optional[bool],
    current_version: Optional[str],
    new_version: Optional[str],
    parse: Optional[str],
    serialize: Optional[List[str]],
    search: Optional[str],
    replace: Optional[str],
    regex: Optional[bool],
    no_configured_files: bool,
    ignore_missing_files: bool,
    ignore_missing_version: bool,
    dry_run: bool,
    commit: Optional[bool],
    tag: Optional[bool],
    sign_tags: Optional[bool],
    tag_name: Optional[str],
    tag_message: Optional[str],
    message: Optional[str],
    commit_args: Optional[str],
) -> None:
    """
    Change the version.

    ARGS may contain any of the following:

    VERSION_PART is the part of the version to increase, e.g. `minor`.
    Valid values include those given in the `--serialize` / `--parse` option.

    FILES are additional file(s) to modify.
    If you want to rewrite only files specified on the command line, use with the
    `--no-configured-files` option.
    """
    setup_logging(verbose)

    logger.info("Starting BumpVersion %s", __version__)

    overrides = get_overrides(
        allow_dirty=allow_dirty,
        current_version=current_version,
        parse=parse,
        serialize=serialize or None,
        search=search,
        replace=replace,
        commit=commit,
        tag=tag,
        sign_tags=sign_tags,
        tag_name=tag_name,
        tag_message=tag_message,
        message=message,
        commit_args=commit_args,
        ignore_missing_files=ignore_missing_files,
        ignore_missing_version=ignore_missing_version,
        regex=regex,
    )

    found_config_file = find_config_file(config_file)
    config = get_configuration(found_config_file, **overrides)
    if args:
        if args[0] not in config.parts.keys():
            raise click.BadArgumentUsage(f"Unknown version part: {args[0]}")
        version_part = args[0]
        files = args[1:]
    else:
        version_part = None
        files = args

    config.allow_dirty = allow_dirty if allow_dirty is not None else config.allow_dirty
    if not config.allow_dirty and config.scm_info.tool:
        config.scm_info.tool.assert_nondirty()

    if no_configured_files:
        config.excluded_paths = list(config.resolved_filemap.keys())

    if files:
        config.add_files(files)
        config.included_paths = files

    logger.dedent()
    do_bump(version_part, new_version, config, found_config_file, dry_run)


@cli.command()
@click.argument("args", nargs=-1, type=str)
@click.option(
    "--config-file",
    metavar="FILE",
    required=False,
    envvar="BUMPVERSION_CONFIG_FILE",
    type=click.Path(exists=True),
    help="Config file to read most of the variables from.",
)
@click.option(
    "-f",
    "--format",
    "format_",
    required=False,
    envvar="BUMPVERSION_FORMAT",
    type=click.Choice(["default", "yaml", "json"], case_sensitive=False),
    default="default",
    help="Config file to read most of the variables from.",
)
@click.option(
    "-i",
    "--increment",
    required=False,
    envvar="BUMPVERSION_INCREMENT",
    type=str,
    help="Increment the version part and add `new_version` to the configuration.",
)
def show(args: List[str], config_file: Optional[str], format_: str, increment: Optional[str]) -> None:
    """Show current configuration information."""
    found_config_file = find_config_file(config_file)
    config = get_configuration(found_config_file)

    if not args:
        do_show("all", config=config, format_=format_, increment=increment)
    else:
        do_show(*args, config=config, format_=format_, increment=increment)


@cli.command()
@click.argument("files", nargs=-1, type=str)
@click.option(
    "--config-file",
    metavar="FILE",
    required=False,
    envvar="BUMPVERSION_CONFIG_FILE",
    type=click.Path(exists=True),
    help="Config file to read most of the variables from.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    required=False,
    envvar="BUMPVERSION_VERBOSE",
    help="Print verbose logging to stderr. Can specify several times for more verbosity.",
)
@click.option(
    "--allow-dirty/--no-allow-dirty",
    default=None,
    required=False,
    envvar="BUMPVERSION_ALLOW_DIRTY",
    help="Don't abort if working directory is dirty, or explicitly abort if dirty.",
)
@click.option(
    "--current-version",
    metavar="VERSION",
    required=False,
    envvar="BUMPVERSION_CURRENT_VERSION",
    help="Version that needs to be updated",
)
@click.option(
    "--new-version",
    metavar="VERSION",
    required=False,
    envvar="BUMPVERSION_NEW_VERSION",
    help="New version that should be in the files. If not specified, it will be None.",
)
@click.option(
    "--parse",
    metavar="REGEX",
    required=False,
    envvar="BUMPVERSION_PARSE",
    help="Regex parsing the version string",
)
@click.option(
    "--serialize",
    metavar="FORMAT",
    multiple=True,
    required=False,
    envvar="BUMPVERSION_SERIALIZE",
    help="How to format what is parsed back to a version",
)
@click.option(
    "--search",
    metavar="SEARCH",
    required=False,
    envvar="BUMPVERSION_SEARCH",
    help="Template for complete string to search",
)
@click.option(
    "--replace",
    metavar="REPLACE",
    required=False,
    envvar="BUMPVERSION_REPLACE",
    help="Template for complete string to replace",
)
@click.option(
    "--regex/--no-regex",
    default=False,
    envvar="BUMPVERSION_REGEX",
    help="Treat the search parameter as a regular expression or explicitly do not treat it as a regular expression.",
)
@click.option(
    "--no-configured-files",
    is_flag=True,
    envvar="BUMPVERSION_NO_CONFIGURED_FILES",
    help=(
        "Only replace the version in files specified on the command line, "
        "ignoring the files from the configuration file."
    ),
)
@click.option(
    "--ignore-missing-version",
    is_flag=True,
    envvar="BUMPVERSION_IGNORE_MISSING_VERSION",
    help="Ignore any Version Not Found errors when searching and replacing in files.",
)
@click.option(
    "--ignore-missing-files",
    is_flag=True,
    envvar="BUMPVERSION_IGNORE_MISSING_FILES",
    help="Ignore any missing files when searching and replacing in files.",
)
@click.option(
    "--dry-run",
    "-n",
    is_flag=True,
    envvar="BUMPVERSION_DRY_RUN",
    help="Don't write any files, just pretend.",
)
def replace(
    files: list,
    config_file: Optional[str],
    verbose: int,
    allow_dirty: Optional[bool],
    current_version: Optional[str],
    new_version: Optional[str],
    parse: Optional[str],
    serialize: Optional[List[str]],
    search: Optional[str],
    replace: Optional[str],
    regex: bool,
    no_configured_files: bool,
    ignore_missing_version: bool,
    ignore_missing_files: bool,
    dry_run: bool,
) -> None:
    """
    Replace the version in files.

    FILES are additional file(s) to modify.
    If you want to rewrite only files specified on the command line, use with the
    `--no-configured-files` option.
    """
    setup_logging(verbose)

    logger.info("Starting BumpVersion %s", __version__)

    overrides = get_overrides(
        allow_dirty=allow_dirty,
        current_version=current_version,
        new_version=new_version,
        parse=parse,
        serialize=serialize or None,
        commit=False,
        tag=False,
        sign_tags=False,
        tag_name=None,
        tag_message=None,
        message=None,
        commit_args=None,
        ignore_missing_version=ignore_missing_version,
        ignore_missing_files=ignore_missing_files,
        regex=regex,
    )

    found_config_file = find_config_file(config_file)
    config = get_configuration(found_config_file, **overrides)

    config.allow_dirty = allow_dirty if allow_dirty is not None else config.allow_dirty
    if not config.allow_dirty and config.scm_info.tool:
        config.scm_info.tool.assert_nondirty()

    if no_configured_files:
        config.excluded_paths = list(config.resolved_filemap.keys())

    if files:
        config.add_files(files)
        config.included_paths = files

    configured_files = [
        ConfiguredFile(file_cfg, config.version_config, search, replace) for file_cfg in config.files_to_modify
    ]

    version = config.version_config.parse(config.current_version)
    if new_version:
        next_version = config.version_config.parse(new_version)
    else:
        next_version = None

    ctx = get_context(config, version, next_version)

    modify_files(configured_files, version, next_version, ctx, dry_run)


@cli.command()
@click.option(
    "--prompt/--no-prompt",
    default=True,
    help="Ask the user questions about the configuration.",
)
@click.option(
    "--destination",
    default="stdout",
    help="Where to write the sample configuration.",
    type=click.Choice(["stdout", ".bumpversion.toml", "pyproject.toml"]),
)
def sample_config(prompt: bool, destination: str) -> None:
    """Print a sample configuration file."""
    if prompt:
        destination = questionary.select(
            "Destination", choices=["stdout", ".bumpversion.toml", "pyproject.toml"], default=destination
        ).ask()

    destination_config = create_configuration(destination, prompt)

    if destination == "stdout":
        print_info(dumps(destination_config))
    else:
        Path(destination).write_text(dumps(destination_config), encoding="utf-8")


@cli.command()
@click.argument("version", nargs=1, type=str, required=False, default="")
@click.option(
    "--config-file",
    metavar="FILE",
    required=False,
    envvar="BUMPVERSION_CONFIG_FILE",
    type=click.Path(exists=True),
    help="Config file to read most of the variables from.",
)
@click.option("--ascii", is_flag=True, help="Use ASCII characters only.")
def show_bump(version: str, config_file: Optional[str], ascii: bool) -> None:
    """Show the possible versions resulting from the bump subcommand."""
    found_config_file = find_config_file(config_file)
    config = get_configuration(found_config_file)
    if not version:
        version = config.current_version
    box_style = "ascii" if ascii else "light"
    visualize(config=config, version_str=version, box_style=box_style)



"""Tests for the ConfiguredFile class."""

from bumpversion.files import ConfiguredFile, FileChange
from bumpversion.version_part import VersionConfig
from bumpversion.versioning.models import VersionComponentSpec


class TestConfiguredFile:
    """Tests for the ConfiguredFile class."""

    class TestClassCreation:
        """Tests for the creation of the ConfiguredFile class."""

        def test_file_change_is_identical_to_input(self):
            """Test that the file change is identical to the input, but not the same object."""
            change = FileChange(
                filename="boobar.txt",
                search="dashes: {current_version}",
                replace="dashes: {new_version}",
                parse=r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                serialize=("{major}-{minor}-{patch}",),
                regex=True,
                ignore_missing_version=False,
                ignore_missing_file=False,
            )
            version_config = VersionConfig(
                parse="(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
                serialize=("{major}.{minor}.{patch}",),
                search="{current_version}",
                replace="{new_version}",
                part_configs={
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                },
            )
            cfg_file = ConfiguredFile(change, version_config)
            assert cfg_file.file_change == change
            assert cfg_file.file_change is not change

        def test_version_config_uses_file_change_attrs(self):
            """Test that the version config uses the file change attributes, and the original part configs."""
            change = FileChange(
                filename="boobar.txt",
                search="dashes: {current_version}",
                replace="dashes: {new_version}",
                parse=r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                serialize=("{major}-{minor}-{patch}",),
                regex=True,
                ignore_missing_version=False,
                ignore_missing_file=False,
            )
            version_config = VersionConfig(
                parse="(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
                serialize=("{major}.{minor}.{patch}",),
                search="{current_version}",
                replace="{new_version}",
                part_configs={
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                },
            )
            expected = VersionConfig(
                parse=change.parse,
                search=change.search,
                replace=change.replace,
                serialize=change.serialize,
                part_configs=version_config.part_configs,
            )
            cfg_file = ConfiguredFile(change, version_config)
            assert cfg_file.version_config == expected



"""Module for creating a new config file."""

from pathlib import Path
from typing import Tuple

import questionary
from tomlkit import TOMLDocument


def create_configuration(destination: str, prompt: bool) -> TOMLDocument:
    """
    Create a new configuration as a TOMLDocument.

    Args:
        destination: `stdout` or a path to a new or existing file.
        prompt: `True` if the user should be prompted for input.

    Returns:
        The TOMLDocument structure with the updated configuration.
    """
    config, destination_config = get_defaults_from_dest(destination)

    if prompt:
        allow_dirty_default = "(Y/n)" if config["allow_dirty"] else "(y/N)"
        answers = questionary.form(
            current_version=questionary.text("What is the current version?", default=config["current_version"]),
            commit=questionary.confirm(
                "Commit changes made when bumping to version control?", default=config["commit"]
            ),
            allow_dirty=questionary.confirm(
                "Allow dirty working directory when bumping?",
                default=config["allow_dirty"],
                instruction=(
                    "If you are also creating or modifying other files (e.g. a CHANGELOG), say Yes. "
                    f"{allow_dirty_default} "
                ),
            ),
            tag=questionary.confirm("Tag changes made when bumping in version control?", default=config["tag"]),
            commit_args=questionary.text(
                "Any extra arguments to pass to the commit command?",
                default=config["commit_args"] or "",
                instruction="For example, `--no-verify` is useful if you have a pre-commit hook. ",
            ),
        ).ask()
        config.update(answers)

    for key, val in config.items():
        destination_config["tool"]["bumpversion"][key] = val if val is not None else ""

    return destination_config


def get_defaults_from_dest(destination: str) -> Tuple[dict, TOMLDocument]:
    """Get the default configuration and the configuration from the destination."""
    from tomlkit import document, parse

    from bumpversion.config import DEFAULTS

    config = DEFAULTS.copy()
    if Path(destination).exists():
        destination_config = parse(Path(destination).read_text(encoding="utf-8"))
    else:
        destination_config = document()

    destination_config.setdefault("tool", {})
    destination_config["tool"].setdefault("bumpversion", {})
    existing_config = destination_config["tool"]["bumpversion"]
    if existing_config:
        config.update(existing_config)

    project_config = destination_config.get("project", {}).get("version")
    config["current_version"] = config["current_version"] or project_config or "0.1.0"
    del config["scm_info"]
    del config["parts"]
    del config["files"]

    return config, destination_config



"""Standard version conventions."""

from bumpversion.versioning.models import VersionComponentSpec, VersionSpec

# Adapted from https://packaging.python.org/en/latest/specifications/version-specifiers/
PEP440_PATTERN = r"""
    v?
    (?:
        (?P<major>0|[1-9]\d*)\.
        (?P<minor>0|[1-9]\d*)\.
        (?P<patch>0|[1-9]\d*)
        (?:                                          # pre-release
            [-_\.]?
            (?P<pre>
                (?P<pre_l>a|b|c|rc|alpha|beta|pre|preview)
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )
        )?
        (?:                                          # post release
            [-_\.]?
            (?P<post>
                post
                [-_\.]?
                [0-9]+
            )
        )?
        (?:                                          # dev release
            [-_\.]?
            (?P<dev>
                dev
                [-_\.]?
                [0-9]+
            )
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?   # local version
"""
PEP440_SERIALIZE_PATTERNS = [
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{post}.{dev}+{local}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{post}.{dev}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{post}+{local}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{dev}+{local}",
    "{major}.{minor}.{patch}.{post}.{dev}+{local}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{post}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}.{dev}",
    "{major}.{minor}.{patch}{pre_l}{pre_n}+{local}",
    "{major}.{minor}.{patch}.{post}.{dev}",
    "{major}.{minor}.{patch}.{post}+{local}",
    "{major}.{minor}.{patch}.{dev}+{local}",
    "{major}.{minor}.{patch}.{pre_l}{pre_n}",
    "{major}.{minor}.{patch}.{post}",
    "{major}.{minor}.{patch}.{dev}",
    "{major}.{minor}.{patch}+{local}",
    "{major}.{minor}.{patch}",
]
PEP440_COMPONENT_CONFIGS = {
    "major": VersionComponentSpec(),
    "minor": VersionComponentSpec(),
    "patch": VersionComponentSpec(),
    "pre_l": VersionComponentSpec(values=["", "a", "b", "rc"]),
    "pre_n": VersionComponentSpec(),
    "post": VersionComponentSpec(depends_on="patch"),
    "dev": VersionComponentSpec(depends_on="patch"),
    "local": VersionComponentSpec(depends_on="patch", optional_value=""),
}


def pep440_version_spec() -> VersionSpec:
    """Return a VersionSpec for PEP 440."""
    return VersionSpec(components=PEP440_COMPONENT_CONFIGS)


# Adapted from https://regex101.com/r/Ly7O1x/3/
SEMVER_PATTERN = r"""
    (?P<major>0|[1-9]\d*)\.
    (?P<minor>0|[1-9]\d*)\.
    (?P<patch>0|[1-9]\d*)
    (?:
        -                             # dash separator for pre-release section
        (?P<pre_l>[a-zA-Z-]+)         # pre-release label
        (?P<pre_n>0|[1-9]\d*)         # pre-release version number
    )?                                # pre-release section is optional
    (?:
        \+                            # plus separator for build metadata section
        (?P<buildmetadata>
            [0-9a-zA-Z-]+
            (?:\.[0-9a-zA-Z-]+)*
        )
    )?                                # build metadata section is optional
"""
SEMVER_SERIALIZE_PATTERNS = [
    "{major}.{minor}.{patch}-{pre_l}{pre_n}+{buildmetadata}",
    "{major}.{minor}.{patch}-{pre_l}{pre_n}",
    "{major}.{minor}.{patch}+{buildmetadata}",
    "{major}.{minor}.{patch}",
]
SEMVER_COMPONENT_CONFIGS = {
    "major": VersionComponentSpec(),
    "minor": VersionComponentSpec(),
    "patch": VersionComponentSpec(),
    "pre_l": VersionComponentSpec(values=["", "a", "b", "rc"]),
    "pre_n": VersionComponentSpec(),
    "buildmetadata": VersionComponentSpec(independent=True),
}


def semver_spec() -> VersionSpec:
    """Return a VersionSpec for SEMVER."""
    return VersionSpec(components=SEMVER_COMPONENT_CONFIGS)



"""Tests of the configuration utilities."""

from pathlib import Path
from typing import Any

import tomlkit
import pytest
from pytest import param

from bumpversion.config.utils import get_all_file_configs, get_all_part_configs, resolve_glob_files
from bumpversion.config.models import FileChange
from bumpversion.config import DEFAULTS
from tests.conftest import inside_dir


def write_config(tmp_path: Path, overrides: dict) -> Path:
    """Write a configuration file."""
    config_file = tmp_path / ".bumpversion.toml"
    defaults = DEFAULTS.copy()
    defaults.pop("parts")
    defaults.pop("scm_info")
    defaults["current_version"] = defaults["current_version"] or "1.2.3"
    defaults["commit_args"] = ""
    config = {"tool": {"bumpversion": {**defaults, **overrides}}}
    config_file.write_text(tomlkit.dumps(config), encoding="utf-8")
    return config_file


class TestGetAllFileConfigs:
    """Tests for the get_all_file_configs function."""

    def test_uses_defaults_for_missing_keys(self, tmp_path: Path):
        """Test the get_all_file_configs function."""
        config_file = write_config(tmp_path, {"files": [{"filename": "setup.cfg"}]})
        config_dict = tomlkit.loads(config_file.read_text())["tool"]["bumpversion"]
        file_configs = get_all_file_configs(config_dict)

        for key in FileChange.model_fields.keys():
            global_key = key if key != "ignore_missing_file" else "ignore_missing_files"
            if key not in ["filename", "glob", "key_path"]:
                file_val = getattr(file_configs[0], key)
                assert file_val == DEFAULTS[global_key]

    @pytest.mark.parametrize(
        ["attribute", "global_value", "file_value"],
        [
            param("parse", DEFAULTS["parse"], r"v(?P<major>\d+)", id="parse"),
            param("serialize", DEFAULTS["serialize"], ("v{major}",), id="serialize"),
            param("search", DEFAULTS["search"], "v{current_version}", id="search"),
            param("replace", DEFAULTS["replace"], "v{new_version}", id="replace"),
            param("ignore_missing_version", True, False, id="ignore_missing_version"),
            param("ignore_missing_files", True, False, id="ignore_missing_files"),
            param("regex", True, False, id="regex"),
        ],
    )
    def test_overrides_defaults(self, tmp_path: Path, attribute: str, global_value: Any, file_value: Any):
        """Configured attributes in the file should override global options."""
        file_attribute = attribute if attribute != "ignore_missing_files" else "ignore_missing_file"
        config_file = write_config(
            tmp_path,
            {
                attribute: global_value,
                "files": [
                    {
                        "filename": "setup.cfg",
                        file_attribute: file_value,
                    }
                ],
            },
        )
        config_dict = tomlkit.loads(config_file.read_text())["tool"]["bumpversion"]
        file_configs = get_all_file_configs(config_dict)
        assert len(file_configs) == 1
        assert file_configs[0].filename == "setup.cfg"
        assert getattr(file_configs[0], file_attribute) == file_value


class TestResolveGlobFiles:
    """Tests for the resolve_glob_files function."""

    def test_all_attributes_are_copied(self, tmp_path: Path):
        """Test that all attributes are copied."""
        file1 = tmp_path.joinpath("setup.cfg")
        file2 = tmp_path.joinpath("subdir/setup.cfg")
        file1.touch()
        file2.parent.mkdir()
        file2.touch()

        file_cfg = FileChange(
            filename=None,
            glob="**/*.cfg",
            key_path=None,
            parse=r"v(?P<major>\d+)",
            serialize=("v{major}",),
            search="v{current_version}",
            replace="v{new_version}",
            ignore_missing_version=True,
            ignore_missing_file=True,
            regex=True,
        )
        with inside_dir(tmp_path):
            resolved_files = resolve_glob_files(file_cfg)

        assert len(resolved_files) == 2
        for resolved_file in resolved_files:
            assert resolved_file.filename.endswith("setup.cfg")
            assert resolved_file.glob is None
            assert resolved_file.key_path is None
            assert resolved_file.parse == r"v(?P<major>\d+)"
            assert resolved_file.serialize == ("v{major}",)
            assert resolved_file.search == "v{current_version}"
            assert resolved_file.replace == "v{new_version}"
            assert resolved_file.ignore_missing_version is True
            assert resolved_file.ignore_missing_file is True
            assert resolved_file.regex is True
